/*  main.c
 *
 *  Main body of xpsk31 application
 */

/*
 *  xpsk31: An application to transmit and receive
 *  PSK31 signals using a computer's sound card
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "main.h"
#include "shared.h"

/* Signal handler */
static void sig_handler( int signal );

/*------------------------------------------------------------------------*/

  int
main (int argc, char *argv[])
{
  /* Command line option returned by getopt() */
  int option;

  /* New and old actions for sigaction() */
  struct sigaction sa_new, sa_old;


  /* Initialize new actions */
  sa_new.sa_handler = sig_handler;
  sigemptyset( &sa_new.sa_mask );
  sa_new.sa_flags = 0;

  /* Register function to handle signals */
  sigaction( SIGINT,  &sa_new, &sa_old );
  sigaction( SIGSEGV, &sa_new, 0 );
  sigaction( SIGFPE,  &sa_new, 0 );
  sigaction( SIGTERM, &sa_new, 0 );
  sigaction( SIGABRT, &sa_new, 0 );

  /* Initialize GTK+ */
  gtk_set_locale ();
  gtk_init (&argc, &argv);

  /*** Process command line options ***/
  while( (option = getopt(argc, argv, "hv") ) != -1 )
	switch( option )
	{
	  case 'h': /* Print usage and exit */
		Usage();
		exit( 0 );

	  case 'v': /* Print version and exit */
		puts( PACKAGE_STRING );
		exit( 0 );

	  default: /* Print usage and exit */
		Usage();
		exit( -1 );

	} /* End of switch( option ) */

  /* Create main window and menu */
  main_window = create_main_window();
  gtk_window_set_title( GTK_WINDOW(main_window), PACKAGE_STRING );
  gbl_popup_menu = create_main_menu();

  /* Get Rx text buffer and scroller */
  gbl_rx_text_buffer = gtk_text_view_get_buffer
	( GTK_TEXT_VIEW(lookup_widget(main_window, "rx_textview")) );
  gbl_rx_scrolledwindow = lookup_widget( main_window, "rx_scrolledwindow" );

  /* Get Tx text buffer and scroller */
  gbl_tx_text_buffer = gtk_text_view_get_buffer
	( GTK_TEXT_VIEW(lookup_widget(main_window, "tx_textview")) );
  gbl_tx_scrolledwindow = lookup_widget( main_window, "tx_scrolledwindow" );

  /* Get scope widget */
  gbl_scope = lookup_widget( main_window, "scope" );
  gbl_scope_label = lookup_widget( main_window, "scope_label" );

  /* Set labels to Xmit and Rcve buttons */
  {
	GtkWidget
	  *button,
	  *xmit_status,
	  *rcve_status;

	button = lookup_widget( main_window, "xmit" );
	xmit_status = gtk_label_new (_("TRANSMIT"));
	gtk_widget_show (xmit_status);
	gtk_container_add (GTK_CONTAINER (button), xmit_status);
	gtk_label_set_justify (GTK_LABEL (xmit_status), GTK_JUSTIFY_LEFT);
	g_object_set_data_full (G_OBJECT (main_window), "xmit_status",
		gtk_widget_ref (xmit_status),
		(GDestroyNotify) gtk_widget_unref);

	button = lookup_widget( main_window, "rcve" );
	rcve_status = gtk_label_new (_("RECEIVE"));
	gtk_widget_show (rcve_status);
	gtk_container_add (GTK_CONTAINER (button), rcve_status);
	gtk_label_set_justify (GTK_LABEL (rcve_status), GTK_JUSTIFY_LEFT);
	g_object_set_data_full (G_OBJECT (main_window), "rcve_status",
		gtk_widget_ref (rcve_status),
		(GDestroyNotify) gtk_widget_unref);
  }

  /* Show main window */
  gtk_widget_show (main_window);

  /* Load runtime config file, abort on error */
  gtk_idle_add( Load_Config, NULL );

  /* Default mode is BPSK */
  SetFlag( MODE_BPSK );
  SetFlag( ENABLE_AFC );

  gtk_main ();

  return 0;
}

/*-------------------------------------------------------------------*/

/*  Load_Config()
 *
 *  Loads the xpsk31rc configuration file
 */

  gboolean
Load_Config( gpointer data )
{
  int
	lb_idx, /* Label buffer index */
	mc_idx, /* Macro buffer index */
	ln_idx, /* Line buffer index  */
	ln_len, /* Line buffer length */
	tg_cnt, /* Count '<' and '>'  */
	idx;

  char
	rc_fpath[64], /* File path to xpsk31rc */
	line[81];     /* Buffer for Load_Line  */

  /* Config file pointer */
  FILE *xpsk31rc;

  /* Set text view font */
  PangoFontDescription *font_desc;

  /* For setting squelch level */
  GtkWidget *hscale;
  GtkAdjustment *adjustment;

  /* ALSA channel names and values */
  char *chan_name[] =
  {
	"FRONT_LEFT",
	"FRONT_RIGHT",
	"REAR_LEFT",
	"REAR_RIGHT",
	"SIDE_LEFT",
	"SIDE_RIGHT",
	"MONO"
  };

  int chan_number[] =
  {
	SND_MIXER_SCHN_FRONT_LEFT,
	SND_MIXER_SCHN_FRONT_RIGHT,
	SND_MIXER_SCHN_REAR_LEFT,
	SND_MIXER_SCHN_REAR_RIGHT,
	SND_MIXER_SCHN_SIDE_LEFT,
	SND_MIXER_SCHN_SIDE_RIGHT,
	SND_MIXER_SCHN_MONO
  };

  int num_chn = 7;

  /* Setup file path to xpsk31rc */
  snprintf( rc_fpath, sizeof(rc_fpath),
	  "%s/xpsk31/xpsk31rc", getenv("HOME") );

  /* Open xpsk31rc file */
  xpsk31rc = fopen( rc_fpath, "r" );
  if( xpsk31rc == NULL )
  {
	perror( rc_fpath );
	Error_Dialog(
		_("Failed to open xpsk31rc file\n"\
		  "Quit xpsk31 and correct"), QUIT );
	return( FALSE );
  }

  /*** Read Sound Card configuration data ***/
  /* Read card name, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Sound Card Name")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.snd_card, line, sizeof(rc_data.snd_card) );

  /* Read pcm device name, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("PCM Device Name")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.pcm_dev, line, sizeof(rc_data.pcm_dev) );

  /* Read DSP rate Samples/sec, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("DSP Rate") ) != SUCCESS )
	return( FALSE );
  rc_data.dsp_rate = atoi( line );
  rc_data.quadrant_4 = rc_data.dsp_rate / AUDIO_FREQUENCY;
  rc_data.quadrant_1 = rc_data.quadrant_4 / 4;
  rc_data.quadrant_2 = rc_data.quadrant_4 / 2;
  rc_data.quadrant_3 = ( 3 * rc_data.quadrant_4 ) / 4;

  /* Read ALSA "channel", abort if EOF */
  if( Load_Line(line, xpsk31rc, _("ALSA Channel") ) != SUCCESS )
	return( FALSE );
  for( idx = 0; idx < num_chn; idx++ )
	if( strcmp(chan_name[idx], line) == 0 )
	  break;
  if( idx == num_chn )
  {
	fclose( xpsk31rc );
	Error_Dialog(
		_("Invalid ALSA channel name\n"\
		  "Quit and correct xpsk31rc"), QUIT);
	return( FALSE );
  }
  rc_data.channel = chan_number[idx];

  /* Set right or left channel buffer index */
  if( strstr(line, _("LEFT")) || strstr(line, _("MONO")) )
	rc_data.use_chn = 0;
  else
	rc_data.use_chn = 1;

  /* Set number of channels */
  if( strstr(line, _("MONO")) )
	rc_data.num_chn = 1;
  else
	rc_data.num_chn = 2;

  /* Read capture source, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Capture Source")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.cap_src, line, sizeof(rc_data.cap_src) );

  /* Read Capture volume control, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Capture Volume Control")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.cap_vol, line, sizeof(rc_data.cap_vol) );

  /* Read Capture volume, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Capture Volume")) != SUCCESS )
	return( FALSE );
  rc_data.cap_lev = atoi( line );

  /* Read Playback master volume control, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Playback Master Volume Control")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.mst_vol, line, sizeof(rc_data.mst_vol) );

  /* Read Playback PCM volume control, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Playback PCM Volume Control")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.pcm_vol, line, sizeof(rc_data.pcm_vol) );

  /* Read Playback volume, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Capture Volume")) != SUCCESS )
	return( FALSE );
  rc_data.pbk_lev = atoi( line );

  /*** End of Sound Card configuration data ***/

  /* Read squelch threshold, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Squelch Threshold") ) != SUCCESS )
	return( FALSE );
  rc_data.sqlch_thr = atoi( line );

  /* Enter squelch level to hscale */
  hscale = lookup_widget( main_window, "squelch_value" );
  adjustment = gtk_range_get_adjustment( GTK_RANGE(hscale) );
  gtk_adjustment_set_value( adjustment, (gdouble)rc_data.sqlch_thr );
  rc_data.sqlch_thr = atoi( line ) << 8;

  /* Read Magniphase reverse flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Reverse Magniphase") ) != SUCCESS )
	return( FALSE );
  rc_data.rev_magniphase = atoi( line );

  /* Read tone duration, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Tone Duration") ) != SUCCESS )
	return( FALSE );
  rc_data.tone_dur = atoi( line );

  /* Read number of reversals, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Number of Reversals") ) != SUCCESS )
	return( FALSE );
  rc_data.num_rev = atoi( line );

  /* Read CW encoder speed, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("CW Encoder Speed") ) != SUCCESS )
	return( FALSE );
  rc_data.wpm = atoi( line );

  /* Read word wrap column, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Word Wrap Column") ) != SUCCESS )
	return( FALSE );
  rc_data.wwrap = atoi( line );

  /* Read font string for text views, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Text View font string") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.font, line, sizeof(rc_data.font) );

  /*** Setup default main menu items ***/

  /* Read CAT serial port device, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("CAT Serial Port") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.cat_serial, line, sizeof(rc_data.cat_serial) );

  /* Read Transceiver type for  CAT, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Transceiver type for CAT") ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "FT847") == 0 )
	rc_data.tcvr_type = FT847;
  else if( strcmp(line, "FT857") == 0 )
	rc_data.tcvr_type = FT857;
  else if( strcmp(line, "K2") == 0 )
	rc_data.tcvr_type = K2;
  else if( strcmp(line, "K3") == 0 )
	rc_data.tcvr_type = K3;
  else if( strcmp(line, _("NONE")) == 0 )
	rc_data.tcvr_type = NONE;
  else
  {
	rc_data.tcvr_type = NONE;
	fclose( xpsk31rc );
	Error_Dialog(
		_("Error reading Transceiver type\n"\
		  "Quit and correct xpsk31rc"), QUIT );
	return( FALSE );
  }

  /* Enable CAT, disable RTS */
  if( rc_data.tcvr_type == NONE )
	ClearFlag( ENABLE_CAT );
  else
  {
	SetFlag( ENABLE_CAT );
	ClearFlag( ENABLE_RTS_DTR );
  }

  /* Read RTS/DTR-Control serial port device, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("RTS-Control Serial Port") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rts_serial, line, sizeof(rc_data.rts_serial) );

  /* Read RTS/DTR-Control choice, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("RTS-Control Enable") ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "DTR") == 0 )
  {
	SetFlag( ENABLE_DTR );
	SetFlag( ENABLE_RTS_DTR );
	ClearFlag( ENABLE_CAT );
  }
  else if( strcmp(line, "RTS") == 0 )
  {
	SetFlag( ENABLE_RTS_DTR );
	ClearFlag( ENABLE_CAT );
  }
  else if( strcmp(line, "NONE") == 0 )
	ClearFlag( ENABLE_RTS_DTR );
  else
  {
	fclose( xpsk31rc );
	Error_Dialog(
		_("Error reading xpsk31rc\n"\
		  "Unrecognized RTS/DTR-Control option\n"\
		  "Quit xpsk31 and correct"), QUIT );
	return( FALSE );
  }

  /* Read Record QSO enable flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Record QSOs Enable") ) != SUCCESS )
	return( FALSE );

  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "record_qsos")), TRUE );
	SetFlag( RECORD_QSO );
  }
  else if( strcmp(line, "no") == 0 )
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "record_qsos")), FALSE );
  else
  {
	fclose( xpsk31rc );
	Error_Dialog(
		_("Error reading xpsk31rc\n"\
		  "Unrecognized menu option\n"\
		  "Quit xpsk31 and correct"), QUIT );
	return( FALSE );
  }

  /* Read Capitalize enable flag, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Capitalize Enable") ) != SUCCESS )
	return( FALSE );

  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "capitalize_letters")), TRUE );
	SetFlag( CAPITAL_LETTERS );
  }
  else if( strcmp(line, "no") == 0 )
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "capitalize_letters")), FALSE );
  else
  {
	fclose( xpsk31rc );
	Error_Dialog(
		_("Error reading xpsk31rc\n"\
		  "Unrecognized menu option\n"\
		  "Quit xpsk31 and correct"), QUIT );
	return( FALSE );
  }

  /* Read operator callsign, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Operator Callsign") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.call, line, sizeof(rc_data.call) );

  /* Read operator name, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Operator Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.name, line, sizeof(rc_data.name) );

  /* Read Zone, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("ITU Zone") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.zone, line, sizeof(rc_data.zone) );

  /* Read QTH, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("QTH Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.qth, line, sizeof(rc_data.qth) );

  /* Read QTH locator, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("QTH Locator") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.loc, line, sizeof(rc_data.loc) );

  /* Read Transmitter name, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Transmitter Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.tx, line, sizeof(rc_data.tx) );

  /* Read Transmitter power, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Transmitter Power") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.tx_power, line, sizeof(rc_data.tx_power) );

  /* Read Transmitter antenna, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Transmitter Antenna") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.tx_ant, line, sizeof(rc_data.tx_ant) );

  /* Read Receiver name, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Receiver Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rx, line, sizeof(rc_data.rx) );

  /* Read Receiver N.Fig, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Receiver Noise Figure") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rx_nfig, line, sizeof(rc_data.rx_nfig) );

  /* Read Receiver antenna, abort if EOF */
  if( Load_Line(line, xpsk31rc, _("Receiver Antenna") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rx_ant, line, sizeof(rc_data.rx_ant) );

  /* Read propagation medium */
  if( Load_Line(line, xpsk31rc, _("Propagation Medium") ) != SUCCESS )
	return( FALSE );
  Strlcpy( qso_record.via, line, sizeof(qso_record.via) );

  /* Read default remarks on QSO */
  if( Load_Line(line, xpsk31rc, _("Default Remarks") ) != SUCCESS )
	return( FALSE );
  Strlcpy( qso_record.remarks, line, sizeof(qso_record.remarks) );

  /*** Read macro labels from rc file ***/
  /* Null buffer pointers just in case */
  for( idx = 0; idx < NUM_OF_LABELS; idx++ )
	macro[idx] = NULL;

  /* Preload a line */
  if( Load_Line(line, xpsk31rc, _("Macro Labels") ) != SUCCESS )
	return( FALSE );

  for( lb_idx = 0; lb_idx < NUM_OF_LABELS; lb_idx++ )
  {
	/* For setting macro button labels */
	GtkLabel *lbl;
	char lname[8];
	char label[LABEL_LENGTH+6];

	/* Abort if first line is not a label */
	if( line[0] != '[' )
	{
	  fprintf( stderr,
		  "xpsk31: First Macro line not a label\n%s\n", line );
	  fclose( xpsk31rc );
	  Error_Dialog(
		  _("Error reading xpsk31rc\n"\
			"First Macro line not a label\n"\
			"Quit xpsk31 and correct"), QUIT);
	  return( FALSE );
	}

	/* Look for closing ']' */
	ln_idx = 1;
	while( (ln_idx <= LABEL_LENGTH) && (line[ln_idx] != ']') )
	  ln_idx++;

	/* Abort if label string > LABEL_LENGTH char + [] */
	if( ln_idx > LABEL_LENGTH )
	{
	  fprintf( stderr,
		  "xpsk31: Macro label too long\n%s\n", line );
	  fclose( xpsk31rc );
	  Error_Dialog(
		  _("Error reading xpsk31rc\n"\
			"Macro label too long\n"\
			"Quit xpsk31 and correct"), QUIT);
	  return( FALSE );
	}

	/* Enter label string if all OK */
	snprintf( label, sizeof(label), "[F%d] ", lb_idx+1 );
	line[strlen(line)-1] = '\0';
	Strlcat( label, &line[1], sizeof(label) );

	/* Set macro button labels */
	snprintf( lname, sizeof(lname), "label%d", lb_idx+1 );
	lbl = GTK_LABEL( lookup_widget(main_window, lname) );
	gtk_label_set_text( lbl, label );

	/*** Read macros ***/
	mc_idx = tg_cnt = 0;
	do
	{
	  if( Load_Line(line, xpsk31rc, _("Macros")) != SUCCESS )
		return( FALSE );

	  /* Read labels if line starts with '[' */
	  if( line[0] == '[' ) break;

	  /* Allocate memory to macro buffer */
	  ln_len = (int)strlen( line );

	  /* Ignore | end-of-macro delimeter */
	  if( line[ln_len-1] == '|' ) ln_len--;

	  if( !mem_realloc((void **)&macro[lb_idx],
			(size_t)(ln_len + mc_idx + 2)) )
		return( FALSE );

	  /* Concatenate lines to macro */
	  for( ln_idx = 0; ln_idx < ln_len; ln_idx++ )
	  {
		macro[lb_idx][mc_idx++] = line[ln_idx];

		/* Count tag delimiters */
		if( (line[ln_idx] == '<') ||
			(line[ln_idx] == '>') )
		  tg_cnt++;
	  }

	  /* Terminate Macro with LF and null */
	  macro[lb_idx][mc_idx++] = LF;
	  macro[lb_idx][mc_idx] = '\0';
	} /* do */
	while( line[ln_len] != '|'  );

	/* Abort if tag delimiters are odd number */
	if( tg_cnt & 1 )
	{
	  fprintf( stderr, "xpsk31: Malformed tag in Macro\n%s\n", line );
	  fclose( xpsk31rc );
	  Error_Dialog(
		  _("Error reading xpsk31rc\n"\
			"Malformed tag in Macro\n"\
			"Quit xpsk31 and correct"), QUIT );
	  return( FALSE );
	}

  } /* for( lb_idx = 0; lb_idx < NUM_OF_LABELS; lb_idx++ ) */

  /*** Do the global config variables ***/
  /* Set up CAT if enabled */
  switch( rc_data.tcvr_type )
  {
	case K2:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, "k2")), TRUE );
	  break;

	case K3:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, "k3")), TRUE );
	  break;

	case FT847:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, "ft847")), TRUE );
	  break;

	case FT857:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			lookup_widget(gbl_popup_menu, "ft857")), TRUE );
	  break;
  }

  /* Set RTS control */
  if( isFlagSet(ENABLE_RTS_DTR) )
  {
	rc_data.tcvr_type = NONE;
	ClearFlag( ENABLE_CAT );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "rts_ctl")), TRUE );
  }

  /* Duration of half of a PSK pulse, in DSP samples */
  rc_data.psk_elem2 = (int)( (double)rc_data.dsp_rate / BAUD_RATE / 2.0 );
  rc_data.psk_elem  = rc_data.psk_elem2 * 2;

  /* Duration of preamble tone in dsp samples */
  rc_data.tone_period = rc_data.dsp_rate / AUDIO_FREQUENCY;

  /* Set volume spinbuttons */
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(lookup_widget(
		  main_window, "rxvol_spinbutton")), rc_data.cap_lev );
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(lookup_widget(
		  main_window, "txvol_spinbutton")), rc_data.pbk_lev );

  /* Puts the tone freq in middle of waterfall */
  idx = (2 * DFT_INPUT_SIZE) / gbl_wfall_width;
  rc_data.dft_stride = rc_data.dsp_rate / AUDIO_FREQUENCY / idx;

  /* Make cosine & sine wave table */
  Make_Cosine_Table();
  Make_Sine_Table();

  /* Set Rx and Tx text view font */
  font_desc = pango_font_description_from_string( rc_data.font );
  gtk_widget_modify_font(
	  GTK_WIDGET(lookup_widget(main_window, "rx_textview")), font_desc );
  gtk_widget_modify_font(
	  GTK_WIDGET(lookup_widget(main_window, "tx_textview")), font_desc );
  pango_font_description_free( font_desc );

  fclose( xpsk31rc );

  return( FALSE );
} /* End of Load_Config() */

/*------------------------------------------------------------------*/

/*  Load_Line()
 *
 *  Loads a line from a file, aborts on failure. Lines beginning
 *  with a '#' are ignored as comments. At the end of file EOF is
 *  returned. Lines assumed maximum 80 characters long.
 */

  int
Load_Line( char *buff, FILE *pfile, char *mesg )
{
  int
	num_chr, /* Number of characters read, excluding lf/cr */
	chr;     /* Character read by getc() */
  char error_mesg[MESG_SIZE];

  /* Prepare error message */
  snprintf( error_mesg, sizeof(error_mesg),
	  _("Error reading %s\n"\
		"Premature EOF (End Of File)"), mesg );

  /* Clear buffer at start */
  buff[0] = '\0';
  num_chr = 0;

  /* Get next character, return error if chr = EOF */
  if( (chr = fgetc(pfile)) == EOF )
  {
	fprintf( stderr, "xpsk31: %s\n", error_mesg );
	fclose( pfile );
	Error_Dialog( error_mesg, QUIT );
	return( EOF );
  }

  /* Ignore commented lines, white spaces and eol/cr */
  while(
	  (chr == '#') ||
	  (chr == ' ') ||
	  (chr == HT ) ||
	  (chr == CR ) ||
	  (chr == LF ) )
  {
	/* Go to the end of line (look for LF or CR) */
	while( (chr != CR) && (chr != LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xpsk31: %s\n", error_mesg );
		fclose( pfile );
		Error_Dialog( error_mesg, QUIT );
		return( EOF );
	  }

	/* Dump any CR/LF remaining */
	while( (chr == CR) || (chr == LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xpsk31: %s\n", error_mesg );
		fclose( pfile );
		Error_Dialog( error_mesg, QUIT );
		return( EOF );
	  }

  } /* End of while( (chr == '#') || ... */

  /* Continue reading characters from file till
   * number of characters = 80 or EOF or CR/LF */
  while( num_chr < 80 )
  {
	/* If LF/CR reached before filling buffer, return line */
	if( (chr == LF) || (chr == CR) ) break;

	/* Enter new character to line buffer */
	buff[num_chr++] = (char)chr;

	/* Get next character */
	if( (chr = fgetc(pfile)) == EOF )
	{
	  /* Terminate buffer as a string if chr = EOF */
	  buff[num_chr] = '\0';
	  return( SUCCESS );
	}

	/* Abort if end of line not reached at 80 char. */
	if( (num_chr == 80) && (chr != LF) && (chr != CR) )
	{
	  /* Terminate buffer as a string */
	  buff[num_chr] = '\0';
	  snprintf( error_mesg, sizeof(error_mesg),
		  _("Error reading %s\n"\
			"Line longer than 80 characters"), mesg );
	  fprintf( stderr, "xpsk31: %s\n%s\n", error_mesg, buff );
	  fclose( pfile );
	  Error_Dialog( error_mesg, QUIT );
	  return( ERROR );
	}

  } /* End of while( num_chr < max_chr ) */

  /* Terminate buffer as a string */
  buff[num_chr] = '\0';

  return( SUCCESS );
} /* End of Load_Line() */

/*------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage( void )
{
  fprintf( stderr, "%s\n",
	  _("Usage: xpsk31 [-hv]") );

  fprintf( stderr, "%s\n",
	  _("       -h: Print this usage information and exit.") );

  fprintf( stderr, "%s\n",
	  _("       -v: Print version number and exit.") );

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/*  Cleanup()
 *
 *  Cleans up before quitting
 */

  void
Cleanup( void )
{
  /* Save outstanding records */
  if( isFlagSet(SAVE_RECORD) )
	Save_QSO_Record();

  /* Close open files */
  if( qso_record_fp != NULL )
  {
	fclose( qso_record_fp );
	qso_record_fp = NULL;
  }

  if( adif_log_fp != NULL )
  {
	fclose( adif_log_fp );
	adif_log_fp = NULL;
  }

  if( station_log_fp != NULL )
  {
	fclose( station_log_fp );
	station_log_fp = NULL;
  }

  /* Release sound card and Tcvr */
  Close_Tcvr_Serial();
  Close_Playback();
  Close_Capture();
  Close_Mixer();

} /* Cleanup( void ) */

/*------------------------------------------------------------------------*/

/*  sig_handler()
 *
 *  Signal Action Handler function
 */

static void sig_handler( int signal )
{
  /* Cleanup and abort */
  Cleanup();

  fprintf( stderr, "\n" );
  switch( signal )
  {
	case SIGINT :
	  fprintf( stderr, "%s\n",  _("xpsk31: Exiting via User Interrupt") );
	  exit( signal );

	case SIGSEGV :
	  fprintf( stderr, "%s\n",  _("xpsk31: Segmentation Fault") );
	  exit( signal );

	case SIGFPE :
	  fprintf( stderr, "%s\n",  _("xpsk31: Floating Point Exception") );
	  exit( signal );

	case SIGTERM :
	  fprintf( stderr, "%s\n",  _("xpsk31: Termination Request received") );
	  exit( signal );

	case SIGABRT :
	  fprintf( stderr, "%s\n",  _("xpsk31: Abort Signal received") );
	  exit( signal );

  } /* switch( signal ) */

} /* End of sig_handler() */

/*------------------------------------------------------------------------*/

/* Functions for testing and setting/clearing flow control flags
 *
 *  See xpsk31.h for definition of flow control flags
 */

/* An int variable holding the single-bit flags */
static int Flags = 0;

  int
isFlagSet( int flag )
{
  return( (Flags & flag) == flag );
}

  int
isFlagClear( int flag )
{
  return( (~Flags & flag) == flag );
}

  void
SetFlag( int flag )
{
  Flags |= flag;
}

  void
ClearFlag( int flag )
{
  Flags &= ~flag;
}

  void
ToggleFlag( int flag )
{
  Flags ^= flag;
}

/*------------------------------------------------------------------------*/
